"""
    Constants for the SDK.
"""

from gatey_sdk.__version__ import __version__ as library_version

# Default API server provider.
# By default, this is API server provided from Florgon,
# but user can override this with self-hosted provider.
API_DEFAULT_SERVER_PROVIDER_URL = "https://api-gatey.florgon.space/v1"

# Expected version from the API server.
API_DEFAULT_SERVER_EXPECTED_VERSION = "0.0.0"

# SDK fields.
SDK_NAME = "gatey.python.official"
SDK_VERSION = library_version
SDK_INFORMATION_DICT = {"sdk.name": SDK_NAME, "sdk.ver": SDK_VERSION}

# Exception attribute names.
EXC_ATTR_SHOULD_SKIP_SYSTEM_HOOK = "gatey_should_skip_system_hook"
EXC_ATTR_WAS_HANDLED = "gatey_was_handled"
EXC_ATTR_IS_INTERNAL = "gatey_is_internal"

# Runtime name for runtime event data.
RUNTIME_NAME = "Python"



# pylint: disable=inconsistent-return-statements
"""
    Stuff for working with exceptions on SDK level.
"""

import sys

from typing import Dict, List, Callable, Optional
from types import TracebackType

from gatey_sdk.exceptions import (
    GateyApiError,
    GateyTransportError,
    GateyTransportImproperlyConfiguredError,
)
from gatey_sdk.consts import (
    EXC_ATTR_SHOULD_SKIP_SYSTEM_HOOK,
    EXC_ATTR_WAS_HANDLED,
)
from gatey_sdk.internal.traceback import (
    get_trace_from_traceback,
    get_variables_from_traceback,
)


def wrap_in_exception_handler(
    *,
    reraise: bool = True,
    exception: Optional[BaseException] = None,
    ignored_exceptions: Optional[List[BaseException]] = None,
    on_catch_exception: Optional[Callable] = None,
    skip_global_handler_on_ignore: bool = False,
) -> Callable:
    """
    Decorator that catches the exception and captures it as Gatey exception.
    :param reraise: If False, will not raise the exception again, application will not fall (WARNING: USE THIS WISELY TO NOT GET UNEXPECTED BEHAVIOR)
    :param exception: Target exception type to capture.
    :param ignored_exceptions: List of exceptions that should not be captured.
    :param on_catch_exception: Function that will be called when an exception is caught.
    :param skip_global_handler_on_ignore: If true, will skip global exception handler if exception was ignored.
    """

    # Default target exception value.
    if exception is None:
        exception = BaseException

    # Default value for ignored exception list (Do not ignore any exceptions)
    if ignored_exceptions is None:
        ignored_exceptions = []

    def decorator(function: Callable):
        def wrapper(*args, **kwargs):
            # pylint: disable=inconsistent-return-statements
            # Gets called when `decorated` function get called.
            try:
                # This will simply return function result if there is no exception.
                return function(*args, **kwargs)
            except exception as e:
                # There is any exception that we should handle occurred.

                # Typed.
                e: BaseException = e

                # Do not handle ignored exceptions.
                if _exception_is_ignored(e, ignored_exceptions):
                    if skip_global_handler_on_ignore:
                        # If we should skip global exception handler.
                        setattr(e, EXC_ATTR_SHOULD_SKIP_SYSTEM_HOOK, True)
                    raise e

                # Call catch event.
                if callable(on_catch_exception):
                    on_catch_exception(e)
                    # Mark as handled.
                    setattr(e, EXC_ATTR_WAS_HANDLED, True)

                # Raise exception again if we expected that.
                if reraise is True:
                    raise e
            return

        return wrapper

    return decorator


def register_system_exception_hook(
    hook: Callable, skip_internal_exceptions: bool = True
):
    """
    Register exception hook for system.
    :param hook: Will be called when exception triggered.
    """

    def _system_exception_hook_handler(
        exception_type: type[BaseException],
        exception: BaseException,
        traceback: TracebackType,
    ):
        # System exception hook handler.

        was_handled = hasattr(exception, EXC_ATTR_WAS_HANDLED)
        if not was_handled and not hasattr(exception, EXC_ATTR_SHOULD_SKIP_SYSTEM_HOOK):
            # If marked as skipped for system hook.
            try:
                # Try to handle this exception with hook.
                hook(exception=exception)
            except (
                GateyApiError,
                GateyTransportError,
                GateyTransportImproperlyConfiguredError,
            ) as e:
                # If there is any error while processing global exception handler.
                if not skip_internal_exceptions:
                    raise e

        # Default system hook.
        sys.__excepthook__(exception_type, exception, traceback)

    # Register system exception hook.
    sys.excepthook = _system_exception_hook_handler


def event_dict_from_exception(
    exception: BaseException, skip_vars: bool = True, include_code_context: bool = True
) -> Dict:
    """
    Returns event dictionary of the event (field) from the raw exception.
    Fetches all required information about system, exception.
    """

    # Get raw exception traceback information.
    exception_traceback = getattr(exception, "__traceback__", None)

    # Query traceback information.
    traceback_vars = get_variables_from_traceback(
        traceback=exception_traceback, _always_skip=skip_vars
    )
    traceback_trace = get_trace_from_traceback(
        exception_traceback, include_code_context=include_code_context
    )

    # Get exception type ("BaseException", "ValueError").
    exception_type = _get_exception_type_name(exception)
    exception_description = str(exception)
    event_dict = {
        "class": exception_type,
        "description": exception_description,
        "vars": traceback_vars,  # Will be migrated to the traceback context later.
        "traceback": traceback_trace,
    }
    return event_dict


def _get_exception_type_name(exception: BaseException) -> str:
    """
    Returns exception type ("BaseException", "ValueError").
    """
    return getattr(type(exception), "__name__", "NoneException")


def _exception_is_ignored(
    exception: BaseException, ignored_exceptions: List[BaseException]
) -> bool:
    """
    Returns True if exception should be ignored based on `ignored_exceptions` list.
    """
    exception_type = type(exception)
    for ignored_exception_type in ignored_exceptions:
        if exception_type == ignored_exception_type:
            return True
    return False



"""
    HTTP Transport. Sends event to the Gatey Server when event sends.
"""
import json
from typing import Optional, Dict
from gatey_sdk.transports.base import BaseTransport
from gatey_sdk.api import Api
from gatey_sdk.auth import Auth
from gatey_sdk.exceptions import GateyTransportImproperlyConfiguredError


class HttpTransport(BaseTransport):
    """
    HTTP Transport. Sends event to the Gatey Server when event sends.
    """

    # Allowed aggreation / composition..
    _api_provider: Api = None
    _auth_provider: Optional[Auth] = None

    def __init__(self, api: Optional[Api] = None, auth: Optional[Auth] = None):
        """
        :param api: Api provider.
        :param auth: Authentication provider.
        """

        BaseTransport.__init__(self)
        self._auth_provider = auth if auth else Auth()
        self._api_provider = api if api else Api(self._auth_provider)
        self._check_improperly_configured()

    @BaseTransport.transport_base_sender_wrapper
    def send_event(self, event_dict: Dict) -> None:
        """
        Sends event to the Gatey API server.
        """
        api_params = self._api_params_from_event_dict(event_dict=event_dict)
        self._api_provider.method("event.capture", send_project_auth=True, **api_params)

    @staticmethod
    def _api_params_from_event_dict(event_dict: Dict[str, str]) -> Dict[str, str]:
        """
        Converts event dict to ready for sending API params dict.
        """
        api_params = {
            "level": event_dict["level"],
        }

        event_params = ["exception", "message", "tags"]

        for param in event_params:
            if param in event_dict:
                api_params[param] = json.dumps(event_dict[param])

        return api_params

    def _check_improperly_configured(self):
        """
        Raises error if auth provider improperly configured for sending event.
        """
        if self._auth_provider.project_id is None:
            raise GateyTransportImproperlyConfiguredError(
                "HttpTransport improperly configured! No project id found in auth provider!"
            )
        if (
            self._auth_provider.server_secret is None
            and self._auth_provider.client_secret is None
        ):
            raise GateyTransportImproperlyConfiguredError(
                "HttpTransport improperly configured! No client / server secret found in auth provider!"
            )



"""
    Void transport. Does nothing, used as test environment.
"""

from gatey_sdk.transports.base import BaseTransport


class VoidTransport(BaseTransport):
    """
    Void transport. Does nothing, used as test environment.
    """

    @BaseTransport.transport_base_sender_wrapper
    def send_event(self, *args, **kwargs) -> None:
        """
        Handles transport event callback (handle event sending).
        Does nothing.
        """



"""
    Package with internal modules, used at core sdk level.

    Provides modules for working with traceback, source code fetching, exceptions at level of sdk (low-level).

    Should not be used by not sdk, as it is not designed to be used from elsewhere.
"""

from gatey_sdk.internal.traceback import (
    get_trace_from_traceback,
    get_variables_from_traceback,
)
from gatey_sdk.internal.source import get_context_lines_from_source_code
from gatey_sdk.internal.exc import (
    wrap_in_exception_handler,
    register_system_exception_hook,
    event_dict_from_exception,
)

__all__ = [
    "get_trace_from_traceback",
    "get_variables_from_traceback",
    "get_context_lines_from_source_code",
    "wrap_in_exception_handler",
    "register_system_exception_hook",
    "event_dict_from_exception",
]



"""
    Print transport. Prints event data, used ONLY as test environment.
"""

import json
from typing import Callable, Any, Dict, Optional, Union
from gatey_sdk.transports.base import BaseTransport


class PrintTransport(BaseTransport):
    """
    Print transport. Prints event data, used ONLY as test environment.
    """

    def __init__(
        self,
        indent: Optional[Union[int, str]] = 2,
        prepare_event: Optional[Callable[[Dict], Dict]] = None,
        print_function: Optional[Callable[[Dict], Any]] = None,
    ):
        """
        :param indent: Indent for json convertion
        :param prepare_event: Function that will be called with event and should return event (can be used for clearing unused data to print)
        :param print_function: Function to pass prepared event data to.
        """
        BaseTransport.__init__(self)
        self._indent = indent
        self._prepare_event = prepare_event if prepare_event else lambda e: e
        self._print_function = print_function if print_function else print

    @BaseTransport.transport_base_sender_wrapper
    def send_event(self, event_dict: Dict) -> None:
        """
        Handles transport event callback (handle event sending).
        Print event data.
        """
        print(
            json.dumps(
                self._prepare_event(event_dict),
                indent=self._indent,
                sort_keys=True,
            )
        )



"""
    Response class.
    Result of all API methods.
    Implements Gatey API response structure (For success).

    If API request will raise error, there will be `gatey_sdk.exceptions.GateyApiError`
"""

from typing import Dict, Any, Optional
from requests import Response as _HttpResponse


class Response:
    """
    Gatey API response structure.
    """

    # Raw response fields.
    _raw_json: Optional[Dict] = None
    _raw_response: Optional[_HttpResponse] = None

    # API response fields.
    _response_version: Optional[str] = None
    _response_object: Optional[Dict] = None  # `success` response field.

    def __init__(self, http_response: _HttpResponse):
        """
        :param http_response: Response object (HTTP).
        """

        # Store raw response to work later.
        self._raw_response = http_response

        # Parse raw response once for working later.
        self._raw_json = self._raw_response.json()
        self._response_object = self._raw_json.get("success", dict())
        self._response_version = self._raw_json.get("v", "-")

    def get(self, key: str, default: Any = None):
        """
        Allows to access Response fields by `response.get(field, default)`.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key: str) -> Any:
        """
        Allows to access Response fields by `response[field]`.
        Notice that this will fall with `KeyError` if field was not found in the response.
        """
        if key not in self._response_object:
            raise KeyError(f"{key} does not exist in the response!")
        field_value = self._response_object.get(key)
        return field_value

    def __getattr__(self, attribute_name: str) -> Any:
        """
        Allows to access Response fields by `response.my_response_var`.
        Notice that this will fall with `AttributeError` if field was not found in the response.
        """
        if attribute_name not in self._response_object:
            raise AttributeError(f"{attribute_name} does not exist in the response!")
        attribute_value = self._response_object.get(attribute_name)
        return attribute_value

    def get_version(self) -> str:
        """
        Returns response API version.
        """
        return self._response_version

    def get_response_object(self) -> Dict:
        """
        Returns response object.
        """
        return self._response_object

    def raw_json(self) -> Dict:
        """
        Returns raw JSON from the response.
        WARNING: Do not use this method.
        """
        return self._raw_json

    def raw_response(self) -> _HttpResponse:
        """
        Returns raw response object.
        WARNING: Do not use this method.
        """
        return self._raw_response



"""
    Base abstract class for all transports.
"""

from typing import Dict, Callable, Any
from gatey_sdk.exceptions import GateyError

# There is need in typing.ParamSpec which is 3.10 feature,
# that is not supported in <=3.9, need better way to type hint decorated method.


# Above public field because of requirements below.
def _transport_base_sender_wrapper(
    func: Callable[[Dict], Any]
) -> Callable[[Dict], bool]:
    """
    Wrapper for transport send event method that converts result to success state (boolean).
    """

    def wrapper(*args, **kwargs) -> bool:
        fail_fast = kwargs.pop("__fail_fast", False)
        try:
            func(*args, **kwargs)
        except GateyError as internal_exception:
            if fail_fast:
                raise internal_exception
            return False
        else:
            return True

    return wrapper


class BaseTransport:
    """
    Base transport class. Cannot be used as transport.
    Abstract class for implementing transport classes.
    """

    def __init__(self):
        pass

    # -> BaseTransport.transport_base_sender_wrapper
    # For inherited.
    @_transport_base_sender_wrapper
    def send_event(self, event_dict: Dict) -> None:
        """
        Handles transport event callback (handle event sending).
        Should be inherited from BaseTransport and implemented in transports.
        """
        raise NotImplementedError()

    @staticmethod
    def transport_base_sender_wrapper(
        func: Callable[[Dict], Any]
    ) -> Callable[[Dict], bool]:
        """
        Wrapper for transports send event methods that converts result to success state.
        """
        return _transport_base_sender_wrapper(func)


__all__ = ["BaseTransport"]



"""
    Main client for Gatey SDK.
    Provides root interface for working with Gatey.
"""

import atexit
from threading import Thread
from time import sleep
from typing import Callable, Union, Dict, List, Optional, Any

# Utils.
from gatey_sdk.utils import (
    get_additional_event_tags,
)
from gatey_sdk.internal.exc import (
    wrap_in_exception_handler,
    register_system_exception_hook,
    event_dict_from_exception,
)


# Components.
from gatey_sdk.api import Api
from gatey_sdk.auth import Auth
from gatey_sdk.transports import build_transport_instance, BaseTransport


class _Client:
    """
    ## Gatey SDK client.
    Main interface for working with Gatey.
    Provides transport, auth, api interfaces.

    ### Example use:
    ```python
    import gatey_sdk
    client = gatey_sdk.Client(...)
    ```
    """

    # Instances.
    transport = None  # Used for sending instance (Passed with aggregation).
    auth = None  # Used for authentication.
    api = None  # Used for sending API HTTP requests.

    # Events data queue that waiting for being sent to server.
    _events_buffer: List[Dict[str, Any]] = []
    _events_buffer_flush_thread: Optional[Thread] = None

    # Settings.

    # If true, will send locals(), globals() variables,
    # alongside with event data.
    exceptions_capture_vars = True

    # If true, will capture source code lines for exception.
    exceptions_capture_code_context = True

    # Buffering settings for bulk sending.
    buffer_events_for_bulk_sending = None
    buffer_events_max_capacity = 0
    buffer_events_flush_every = 10.0

    # Include event data.
    include_runtime_info = True
    include_platform_info = True
    include_sdk_info = True

    # Default tags.
    default_tags_context = dict()

    def __init__(
        self,
        *,
        transport: Optional[Union[BaseTransport, Callable]] = None,
        # Settings.
        global_handler_skip_internal_exceptions: bool = True,
        buffer_events_for_bulk_sending: bool = False,
        buffer_events_max_capacity: int = 3,
        buffer_events_flush_every: float = 10.0,
        handle_global_exceptions: bool = False,
        include_runtime_info: bool = True,
        include_platform_info: bool = True,
        include_sdk_info: bool = True,
        exceptions_capture_vars: bool = False,
        exceptions_capture_code_context: bool = True,
        # User auth settings.
        access_token: Optional[str] = None,
        # SDK auth settings.
        project_id: Optional[int] = None,
        server_secret: Optional[str] = None,
        client_secret: Optional[str] = None,
        check_api_auth_on_init: bool = True,
    ):
        """
        :param transport: BaseTransport layer for sending event to the server / whatever else.
        :param global_handler_skip_internal_exceptions:
        :param buffer_events_for_bulk_sending: Will buffer all events (not send immediatly) and will do bulk send when this is required (at exit, or when reached buffer max cap)
        :param buffer_events_max_capacity: Maximal size of buffer to do bulk sending (left 0 for no cap).
        :param handle_global_exceptions: Will catch all exception (use system hook for that).
        :param include_runtime_info: If true, will send runtime information.
        :param include_platform_info: If true will send platform information.
        :param include_sdk_info: If true will send SDK information.
        :param exceptions_capture_vars: Will capture variable (globals, locals) for all exceptions.
        :param exceptions_capture_code_context: Will capture source code context (lines).
        :param access_token: User access token for calling API as authorized user (not for catching events).
        :param project_id: ID of the project from Gatey dashboard.
        :param server_secret: From Gatey dashboard.
        :param client_secret: From Gatey dashboard.
        :param check_api_auth_on_init: Will do hard auth check at init.
        """

        # Components.
        self.auth = Auth(
            access_token=access_token,
            project_id=project_id,
            server_secret=server_secret,
            client_secret=client_secret,
        )
        self.api = Api(auth=self.auth)
        self.transport = build_transport_instance(
            transport_argument=transport, api=self.api, auth=self.auth
        )

        # Options.
        self.exceptions_capture_vars = exceptions_capture_vars
        self.exceptions_capture_code_context = exceptions_capture_code_context
        self.buffer_events_for_bulk_sending = buffer_events_for_bulk_sending
        self.buffer_events_max_capacity = buffer_events_max_capacity
        self.buffer_events_flush_every = buffer_events_flush_every
        self.include_runtime_info = include_runtime_info
        self.include_platform_info = include_platform_info
        self.include_sdk_info = include_sdk_info

        # Tags like platform, sdk, etc.
        self.default_tags_context = self._build_default_tags_context(
            foreign_tags=dict()
        )

        # Check API auth if requested and should.
        # Notice that auth check is not done when you are using custom transports.
        # (even it is default transport)
        if check_api_auth_on_init is True and transport is None:
            self.api.do_hard_auth_check()

        # Register system hooks.
        self._bind_system_exit_hook()
        if handle_global_exceptions is True:
            register_system_exception_hook(
                hook=self._on_catch_exception_hook,
                skip_internal_exceptions=global_handler_skip_internal_exceptions,
            )

        # Start flush thread for bulk sending after timeout (every N).
        if self.buffer_events_for_bulk_sending:
            self._ensure_running_buffer_flush_thread()

    def catch(
        self,
        *,
        reraise: bool = True,
        exception: Optional[BaseException] = None,
        ignored_exceptions: Optional[List[BaseException]] = None,
        skip_global_handler_on_ignore: bool = False,
    ):
        """
        Decorator that catches the exception and captures it as Gatey exception.
        :param reraise: If False, will not raise the exception again, application will not fall (WARNING: USE THIS WISELY TO NOT GET UNEXPECTED BEHAVIOR)
        :param exception: Target exception type to capture.
        :param ignored_exceptions: List of exceptions that should not be captured.
        :param skip_global_handler_on_ignore: If true, will skip global exception handler if exception was ignored.
        """
        return wrap_in_exception_handler(
            reraise=reraise,
            exception=exception,
            ignored_exceptions=ignored_exceptions,
            skip_global_handler_on_ignore=skip_global_handler_on_ignore,
            on_catch_exception=self._on_catch_exception_hook,
        )

    def capture_event(
        self,
        event: Dict,
        level: str,
        tags: Optional[Dict[str, str]] = None,
        include_default_tags: bool = True,
    ) -> bool:
        """
        Captures raw event data and passes it to the transport.
        You should not use this function directly, please use `capture_message` or `capture_exception`!
        This function is used as low-level call to capture all events.

        :param event: Raw event dictionary that will be updated with base event data (including tags).
        :param level: Level of the event.
        :param tags: Dictionary of the tags (string-string).
        :param include_default_tags: If false, will force to not pass default tags context of the client to the event.
        """
        if tags is None or not isinstance(tags, Dict):
            tags = dict()

        if not isinstance(level, str):
            raise TypeError("Level of the event should be always string!")
        if not isinstance(event, Dict):
            raise TypeError("Event data should be Dict!")

        # Include default tags if requred.
        tags.update(self.default_tags_context if include_default_tags else {})

        # Build event data.
        event_dict = event.copy()
        event_dict["tags"] = tags
        event_dict["level"] = level.lower()

        # Will buffer or immediatly send event.
        return self._buffer_captured_event(event_dict=event_dict)

    def capture_message(
        self,
        message: str,
        level: str = "info",
        *,
        tags: Optional[Dict[str, str]] = None,
        include_default_tags: bool = True,
    ) -> bool:
        """
        Captures message event.
        :param level: String of the level (INFO, DEBUG, etc)
        :param message: Message string.
        :param tags: Dictionary of the tags (string-string).
        :param include_default_tags: If false, will force to not pass default tags context of the client to the event.
        """
        event_dict = {"message": message}
        return self.capture_event(
            event=event_dict,
            level=level,
            tags=tags,
            include_default_tags=include_default_tags,
        )

    def capture_exception(
        self,
        exception: BaseException,
        *,
        level: str = "error",
        tags: Optional[Dict[str, str]] = None,
        include_default_tags: bool = True,
    ) -> bool:
        """
        Captures exception event.
        :param exception: Raw exception.
        :param level: Level of the event that will be sent.
        :param tags: Dictionary of the tags (string-string).
        :param include_default_tags: If false, will force to not pass default tags context of the client to the event.
        """
        exception_dict = event_dict_from_exception(
            exception=exception,
            skip_vars=not self.exceptions_capture_vars,
            include_code_context=self.exceptions_capture_code_context,
        )
        event_dict = {"exception": exception_dict}
        if "description" in exception_dict:
            event_dict["message"] = exception_dict["description"]
        return self.capture_event(
            event=event_dict,
            level=level,
            tags=tags,
            include_default_tags=include_default_tags,
        )

    def update_default_tag(self, tag_name: str, tag_value: str) -> None:
        """
        Updates default value for tag.
        """
        if not isinstance(tag_value, str) or not isinstance(tag_name, str):
            raise TypeError("Tag name and value should be strings!")
        self.default_tags_context[tag_name] = tag_value

    def bulk_send_buffered_events(self) -> bool:
        """
        Sends all buffered events.
        Returns is all events was sent.
        """

        # Copy buffer and clear it.
        events_bulk_queue = self._events_buffer.copy()
        self._events_buffer = []

        for event_dict in events_bulk_queue:
            if not self.transport.send_event(event_dict=event_dict):
                # If failed, pass back.
                self._events_buffer.append(event_dict)
        return len(self._events_buffer) == 0

    def force_drop_buffered_events(self) -> None:
        """
        Drops (removes) buffered events explicitly.
        """
        self._events_buffer = []

    def _build_default_tags_context(
        self, foreign_tags: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Returns default tags dict (context).
        """
        default_tags = dict()
        default_tags = get_additional_event_tags(
            include_runtime_info=self.include_runtime_info,
            include_platform_info=self.include_runtime_info,
            include_sdk_info=self.include_sdk_info,
        )
        default_tags.update(foreign_tags)
        return default_tags

    def _ensure_running_buffer_flush_thread(self) -> Thread:
        """
        Runs buffer flush thread is it not running, and returns thread.
        Flush thread is used to send events buffer after some time, not causing to wait core application
        for terminate or new events that will trigger bulk sending (buffer flush).
        """

        if (
            isinstance(self._events_buffer_flush_thread, Thread)
            and self._events_buffer_flush_thread.is_alive()
        ):
            # If thread is currently alive - there is no need to create new thread by removing old reference.
            return self._events_buffer_flush_thread

        # Thread.
        self._events_buffer_flush_thread = Thread(
            target=self._buffer_flush_thread_target,
            args=(),
            name="gatey_sdk.events_buffer.flusher",
        )

        # Mark thread as daemon (which is required for graceful main thread termination) and start.
        self._events_buffer_flush_thread.daemon = True
        self._events_buffer_flush_thread.start()

        return self._events_buffer_flush_thread

    def _buffer_flush_thread_target(self) -> None:
        """
        Thread target for events buffer flusher.
        """
        while True:
            sleep(self.buffer_events_flush_every)
            self.bulk_send_buffered_events()

    def _on_catch_exception_hook(self, exception):
        """
        Hook that will be called when catched exception.
        (except via `capture_exception`)
        """
        return self.capture_exception(exception=exception)

    def _bind_system_exit_hook(self) -> None:
        """
        Binds system hook for exit.
        """

        # Bulk send buffered events at exit.
        if self.buffer_events_for_bulk_sending:
            atexit.register(self.bulk_send_buffered_events)

    def _events_buffer_is_full(self) -> bool:
        """
        Returns is events buffer is full and should be bulk sent.
        """
        if (
            self.buffer_events_max_capacity is None
            or self.buffer_events_max_capacity == 0
        ):
            return False
        return len(self._events_buffer) >= self.buffer_events_max_capacity

    def _buffer_captured_event(self, event_dict: Dict) -> bool:
        """
        Buffers captured event for bulk sending later.
        """

        # Pass directly if should not buffer.
        if not self.buffer_events_for_bulk_sending:
            return self.transport.send_event(event_dict=event_dict, __fail_fast=True)

        # Do buffer and send if required.
        self._events_buffer.append(event_dict)
        if self._events_buffer_is_full():
            return self.bulk_send_buffered_events()
        return True


Client = _Client



# pylint: disable=arguments-differ
# pylint: disable=raise-missing-from
"""
    Function transport. Calls your function when event sends.
"""

from typing import Callable, Dict, Any
from gatey_sdk.transports.base import BaseTransport
from gatey_sdk.exceptions import GateyTransportError


class FuncTransport(BaseTransport):
    """
    Function transport. Calls your function when event sends.
    """

    skip_to_internal_exception: bool = False
    _function: Callable[..., Any]

    def __init__(self, func: Callable, *, skip_to_internal_exception: bool = False):
        """
        :param function: Function to call when event sends.
        """
        BaseTransport.__init__(self)
        self.skip_to_internal_exception = skip_to_internal_exception
        self._function = func

    @BaseTransport.transport_base_sender_wrapper
    def send_event(self, event_dict: Dict) -> None:
        """
        Handles transport event callback (handle event sending).
        Function transport just takes event and passed it raw to function call.
        """
        if self.skip_to_internal_exception:
            try:
                self._function(event_dict)
            except Exception as _:
                raise GateyTransportError(
                    "Unable to handle event send with Function transport (FuncTransport)."
                )
            return
        self._function(event_dict)



"""
    Transports for Client.
"""

from typing import Any, Union, Optional

from gatey_sdk.transports.base import BaseTransport
from gatey_sdk.transports.http import HttpTransport
from gatey_sdk.transports.func import FuncTransport
from gatey_sdk.transports.void import VoidTransport
from gatey_sdk.transports.print import PrintTransport

from gatey_sdk.exceptions import (
    GateyTransportImproperlyConfiguredError,
)
from gatey_sdk.api import Api
from gatey_sdk.auth import Auth


def build_transport_instance(
    transport_argument: Any = None,
    api: Optional[Api] = None,
    auth: Optional[Auth] = None,
) -> Union[BaseTransport, None]:
    """
    Builds transport instance by transport argument.
    """
    transport_class = None

    if transport_argument is None:
        # If nothing is passed, should be default http transport type.
        return HttpTransport(api=api, auth=auth)

    if isinstance(transport_argument, type) and issubclass(
        transport_argument, BaseTransport
    ):
        # Passed subclass (type) of BaseTransport as transport.
        # Should be instantiated as cls.
        transport_class = transport_argument
        if transport_class in (VoidTransport, PrintTransport):
            return transport_class()
        try:
            return transport_class(api=api, auth=auth)
        except TypeError as _transport_params_error:
            raise GateyTransportImproperlyConfiguredError(
                "Failed to build transport instance. Please instantiate before or except your transport to handle `api`, `auth` params in constructor!"
            ) from _transport_params_error

    if isinstance(transport_argument, BaseTransport):
        # Passed already constructed transport, should do nothing.
        return transport_argument

    if callable(transport_argument):
        # Passed callable (function) as transport.
        # Should be Function transport, as it handles raw function call.
        return FuncTransport(func=transport_argument)

    # Unable to instantiate transport instance.
    raise GateyTransportImproperlyConfiguredError(
        "Failed to build transport instance. Please pass valid transport argument!"
    )


# Base transport should be used as typing
# (polymorphism for expecting any implementation of abstract class as interface)
# You are not supposed to implement own class-transports (as there is always no need for that)
# use FuncTransport with your function, to make work done with native library implementation.
__all__ = [
    # Typing.
    "BaseTransport",
    # Default.
    "HttpTransport",
    # Debug.
    "FuncTransport",
    "PrintTransport",
    "VoidTransport",
]
